/// RefCell<T>
/// 相较于Rc其和Box一样只能有一个拥有者，但是其可以修改内部的值，Rc只能读.
pub trait Messenger {
    fn send(&self, msg: &str);
}
use std::cell::RefCell;
pub struct LimitTracker<'a, T: 'a + Messenger> {
    messenger: &'a T,
    value: usize,
    max: usize,
}

/// 为所有实现了messenger trait的类型提供一个限制检测的功能，初始化时指定最大值
impl<'a, T> LimitTracker<'a, T>
where
    T: Messenger,
{
    pub fn new(messenger: &T, max: usize) -> LimitTracker<T> {
        LimitTracker {
            messenger,
            value: 0,
            max,
        }
    }
    /// 这是要测试的方法，该方法，根据初始化时定义的max值和当前值的的比值，
    /// 当超过指定的比利时候发生不同的消息，没有返回值
    /// 其可以被所有实现了Messenger trait的类型使用。
    pub fn set_value(&mut self, value: usize) {
        self.value = value;
        let percentage_of_max = self.value as f64 / self.max as f64;

        if percentage_of_max >= 1.0 {
            self.messenger.send("超过限额")
        } else if percentage_of_max >= 0.9 {
            self.messenger.send("已使用90%")
        } else if percentage_of_max >= 0.75 {
            self.messenger.send("已使用75%")
        }
    }
}

#[cfg(test)]
mod tests {
    use super::*;

    /// 为了测试上面的LimitTracker的set_value方法这里制作一个模拟对象进行测试，
    /// 因为set_value本身不返回任何返回值就没法使用assert_eq等进行检测，所以
    /// 这里设计一个模拟对象，并实现send方法，在set_value函数时将发送的字符串存储起来
    /// 这样跑完set_value之后我们可以通过模拟对象里面的数据来判断是否符合预期
    /// 从而来获知set_value是否能够正常工作
    /// 但是Messenger trait的send方法的参数是不可变引用，那么在send函数里面是不可以修改成员的变量的
    /// 这时候就可以使用RefCell来获得不可变引用的可变
    struct MockMessenger {
        sent_messenger: RefCell<Vec<String>>,
    }

    impl MockMessenger {
        fn new() -> MockMessenger {
            MockMessenger {
                sent_messenger: RefCell::new(vec![]), // 使用RefCell存储数据
            }
        }
    }
    impl Messenger for MockMessenger {
        fn send(&self, msg: &str) {
            // borrow_mut 获得数据的可变引用
            self.sent_messenger.borrow_mut().push(String::from(msg));
        }
    }

    #[test]
    fn it_sends_an_over_75_percent_wrning_messager() {
        let obj = MockMessenger::new();
        let mut limit_tracker = LimitTracker::new(&obj, 100);

        limit_tracker.set_value(80);
        // borrow_mut 获得数据的不可变引用
        assert_eq!(obj.sent_messenger.borrow().len(), 1);
    }
}
