/* 我们设计一个summary trait我们希望这个trait可以提供一个摘要的功能，
   比如我们的博客，微博，文章等都需要一个摘要，方便在信息流展示时进行展示。
*/
pub trait Summary {
    //定义一个summary trait下的方法 summerize，该方法返回一个摘要的字符串
    fn summarize(&self) -> String;
    /* trait 定义得到每个方法在使用了impl trait的数据结构中都需要完整实现所有的方法 */
    //fn reply(&self) -> bool; // 比如定义了这个方法后，下面引入了Summary的NewsArticle和Tweet都需要进行实现！

    // 定义一个author方法，并将其实现，这就是trait的默认方法
    // 如果引入了summary的数据结构没有实现该方法，那么在调用时就会调用该默认方法。
    fn author(&self) -> String {
        String::from("unkown")
    }
}

// 定义一个新闻文章的数据结构，
pub struct NewsArticle {
    pub headline:String, // 标题
    pub location:String, // 地点
    pub author: String, // 作者
    pub content:String, // 内容
}
//定义一个推文的数据结构
pub struct Tweet {
    pub username:String,//用户名
    pub content:String, // 内容
    pub reply: bool, //是否可回复
    pub retweet:bool, //是否可转发
}


//为 NewsArticle 实现 summary
// 语法为 impl <trait name> for <data_type>
impl Summary for NewsArticle {
    fn summarize(&self) -> String {
        //对于新闻来说我们使用标题，作者和地点。生成一个摘要信息
        format!("{}, by {} ({})",self.headline, self.author(), self.location)
    }

    //覆盖掉Summary的默认author实现。在NewsArticle 实例调用该方法时只能调用到该函数，
    //并且该函数无法再调用到默认的author函数
    fn author(&self) -> String {
        let mut one = String::new();
        self.author.clone_into(&mut one);
        one
    }
}

impl Summary for Tweet {
    fn summarize(&self) -> String {
        //对于推文来说我们使用作者和推文内容作为摘要
        format!("{}: {}", self.author(), self.content)
    }
}

/* 经过上面针对两种不同的数据结构对于summery的实现，
   我们可以使得他们都有一种共同的功能-> 生成摘要
 */

// 通过引入Dsiplay trait来使得本地的数据结构Tweet具备 Display的功能
use std::{fmt, fmt::{Display, Formatter}};
impl Display for Tweet {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        //实现Display trait的功能
        write!(f, "{}:\n {} \n reply:{} retweet:{}", self.username, self.content, self.reply, self.retweet)
    }
}

/*
  我们指定的函数参数时，通常需要指定参数的类型。只有对应类型的数据才可以传入，
  RUST提供了另外一种限制参数类型的方式：impl trait。
  通过这种定义形参的方法可以使得只有实现了指定trait的类型才可以使用该函数，但是又不限制具体的数据类型，
  这种使用方式适合针对某个trait所做的功能，使得所有实现该trait的类型都可以免于实现类似的函数，减少代码。
  var: impl xxx
 */
//定义一个通知函数。只有具备summary trait的类型才可以作为参数传递进来。
//这样的一个通用的通知函数，使得不用为每个想要通知的数据类型都写一个类似的函数
pub fn notify(not: impl Summary){
    println!("new notification: [{}]", not.summarize());
}

// 传入两个参数都用impl Summary修饰，表示not1和not2都是实现了Summary的类型，但是他们不一定是同一个数据类型。
pub fn notify1(not1: impl Summary, not2: impl Summary){
    println!("new notification1: [{}]", not1.summarize());
    println!("new notification1: [{}]", not2.summarize());
}

// 传入两个参数都都是类型T，T类型要求是符合Summary trait的类型,
// 这里要注意：not1和not2都必须是同一种类型才可以
pub fn notify2<T: Summary>(not1: T, not2: T){
    println!("new notification2: [{}]", not1.summarize());
    println!("new notification2: [{}]", not2.summarize());
}

//要求参数同时满足多个trait的类型限制
pub fn some_func0<T: Summary+Copy>(not1: T){
}

//多个参数，每个参数各自限制不用的trait限制
pub fn some_func<T: Summary+Copy, U: Summary + Display>(not1: T, not2: U){
}
//简化的写法使用Where来将限制提在函数签名外，使得函数名和参数列表比较靠近，where可以换行，多行来写
pub fn some_func1<T, U>(not1: T, not2: U) -> i32
    where T: Summary+Copy, U: Summary + Display
{
    0
}

// 返回值同样可以限制为实现了某个trait的数据类型, 也可以多种trait限制叠加。
// 返回一个实现了Summary trait和Copy类型的变量
pub fn some_func2<T, U>(not1: T, not2: U) -> impl Summary+Copy
    where T: Summary+Copy, U: Summary + Display
{
    not1
}